/* GStreamer
 * Copyright (C) <2001> Steve Baker <stevebaker_org@yahoo.co.uk>
 *
 * float2int.c
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gstfloat2int.h>


static GstElementDetails float2int_details = {
  "Float to Integer effect",
  "Filter/Effect",
  "Convert from floating point to integer audio data",
  VERSION,
  "Steve Baker <stevebaker_org@yahoo.co.uk>",
  "(C) 2001",
};


enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_PAN,
};

GST_PAD_TEMPLATE_FACTORY (float2int_sink_factory,
  "sink%d",
  GST_PAD_SINK,
  GST_PAD_REQUEST,
  GST_CAPS_NEW (
    "float2int_sink",
    "audio/raw",
    "rate",       GST_PROPS_INT_RANGE (4000, 96000),
    "format",     GST_PROPS_STRING ("float"),
    "layout",     GST_PROPS_STRING ("gfloat"),
    "intercept",  GST_PROPS_FLOAT(0.0),
    "slope",      GST_PROPS_FLOAT(1.0),
    "channels",   GST_PROPS_INT (1)
  )
);

GST_PAD_TEMPLATE_FACTORY (float2int_src_factory,
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_CAPS_NEW (
    "float2int_src",
    "audio/raw",
    "format",     GST_PROPS_STRING ("int"),
    "channels",   GST_PROPS_INT_RANGE (1, G_MAXINT),
    "rate",       GST_PROPS_INT_RANGE (4000, 96000),
    "law",        GST_PROPS_INT (0),
    "endianness", GST_PROPS_INT (G_BYTE_ORDER),
    "width",      GST_PROPS_INT (16),
    "depth",      GST_PROPS_INT (16),
    "signed",     GST_PROPS_BOOLEAN (TRUE)
  )
);

static GstPadTemplate *srctempl, *sinktempl;

static void	gst_float2int_class_init		(GstFloat2IntClass *klass);
static void	gst_float2int_init			(GstFloat2Int *plugin);

static GstPadConnectReturn gst_float2int_connect	(GstPad *pad, GstCaps *caps);

static void	gst_float2int_set_property		(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void	gst_float2int_get_property		(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);

static GstPad*  gst_float2int_request_new_pad (GstElement *element, GstPadTemplate *temp, const
                                               gchar *unused);

static void	gst_float2int_chain_gint16		(GstPad *pad, GstBuffer *buf_in);
static inline GstFloat2Int* gst_float2int_get_plugin(GstPad *pad,GstBuffer *buf);

static GstElementClass *parent_class = NULL;

GType
gst_float2int_get_type(void) {
  static GType float2int_type = 0;

  if (!float2int_type) {
    static const GTypeInfo float2int_info = {
      sizeof(GstFloat2IntClass),      NULL,
      NULL,
      (GClassInitFunc)gst_float2int_class_init,
      NULL,
      NULL,
      sizeof(GstFloat2Int),
      0,
      (GInstanceInitFunc)gst_float2int_init,
    };
    float2int_type = g_type_register_static(GST_TYPE_ELEMENT, "GstFloat2Int", &float2int_info, 0);
  }
  return float2int_type;
}

/* this is a very brittle function. it's hard to come up with a "correct
   behavior". what I (wingo) have chosen is as follows:
   * if the rate of the float2int plugin has already been set by a previous connection
     * if the new rate is different from the old rate, bail
     * otherwise return 'ok' (don't set other caps, no need -- only the rate needs
       setting, and it's the same)
   * otherwise set all of the caps (except for on the connecting pad, of course)
     * bail if that fails, possibly leaving the plugin in an inconsistent state

   FIXME: we really should do something about the inconsistent state case,
   possibly disconnecting the offending pad. the assumption with this element,
   though, is that the pipelines on the float side will be parallel (more or
   less the same) and so we should never run into that case. that might be a
   programmer error more than anything.
 */
static GstPadConnectReturn
gst_float2int_connect (GstPad *pad, GstCaps *caps)
{
  GstFloat2Int *filter;
  GstCaps *intcaps, *floatcaps;
  GSList *l;
  
  filter = GST_FLOAT2INT (GST_PAD_PARENT (pad));

  /* we have two variables to play with: "rate" and "channels" on the int side. */
  if (GST_CAPS_IS_FIXED (caps)) {
    if (pad == filter->srcpad) {
      /* it's the int pad */
      if (!filter->intcaps) {
	gint rate;

        floatcaps = gst_caps_copy (gst_pad_template_get_caps (float2int_sink_factory ()));
        /* FIXME: we really shouldn't have to convert to int and back */
        gst_caps_get_int (caps, "rate", &rate);
        gst_caps_set (floatcaps, "rate", GST_PROPS_INT (rate));
        /* we now know that the caps are fixed. let's set them. this is a hack
           but oh well. */
        floatcaps->fixed = TRUE;
        
        for (l=filter->sinkpads; l; l=l->next)
          if (!gst_pad_try_set_caps ((GstPad*) l->data, floatcaps))
            return GST_PAD_CONNECT_REFUSED;
        
        /* FIXME: refcounting? */
        filter->floatcaps = floatcaps;
        filter->intcaps = caps;
        gst_caps_get_int (caps, "channels", &filter->channels);
        return GST_PAD_CONNECT_OK;
      } else if (gst_caps_intersect (caps, filter->intcaps)) {
        /* they have to be the same since both are fixed */
        return GST_PAD_CONNECT_OK;
      } else {
        return GST_PAD_CONNECT_REFUSED;
      }
    } else {
      /* it's a float pad */
      if (!filter->intcaps) {
	gint rate;

        /* FIXME: we really shouldn't have to convert to int and back */
        intcaps = gst_caps_copy (gst_pad_template_get_caps (float2int_src_factory ()));
        gst_caps_get_int (caps, "rate", &rate);
        gst_caps_set (intcaps, "rate", GST_PROPS_INT (rate));
        gst_caps_set (intcaps, "channels", GST_PROPS_INT (filter->numsinkpads));
        intcaps->fixed = TRUE;
        gst_caps_debug (intcaps, "int source pad caps going into try_set_caps()");
        
        if (!gst_pad_try_set_caps (filter->srcpad, intcaps))
          return GST_PAD_CONNECT_REFUSED;
        for (l=filter->sinkpads; l; l=l->next)
          if (! ((GstPad*) l->data == pad || gst_pad_try_set_caps ((GstPad*) l->data, caps)))
            return GST_PAD_CONNECT_REFUSED;
        
        /* FIXME: refcounting? */
        filter->floatcaps = caps;
        filter->intcaps = intcaps;
        filter->channels = filter->numsinkpads;
        return GST_PAD_CONNECT_OK;
      } else if (gst_caps_intersect (caps, filter->floatcaps)) {
        /* they have to be the same since both are fixed */
        return GST_PAD_CONNECT_OK;
      } else {
        return GST_PAD_CONNECT_REFUSED;
      }
    }
  } else {
    return GST_PAD_CONNECT_DELAYED;
  }
}

static void
gst_float2int_class_init (GstFloat2IntClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*)klass;
  gstelement_class = (GstElementClass*)klass;

  parent_class = g_type_class_ref(GST_TYPE_ELEMENT);

  gobject_class->set_property = gst_float2int_set_property;
  gobject_class->get_property = gst_float2int_get_property;

  gstelement_class->request_new_pad = gst_float2int_request_new_pad;
}

static void
gst_float2int_init (GstFloat2Int *plugin)
{
  plugin->srcpad = gst_pad_new_from_template(srctempl,"src");
  gst_element_add_pad(GST_ELEMENT(plugin),plugin->srcpad);
  gst_pad_set_connect_function(plugin->srcpad, gst_float2int_connect);

  plugin->numsinkpads = 0;
  plugin->numbuffers = 0;
  plugin->sinkpads = NULL;
  plugin->buffers = NULL;
  plugin->minbuffersize=G_MAXINT;

  plugin->channels=1;
}

static GstPad*
gst_float2int_request_new_pad (GstElement *element, GstPadTemplate *templ, const gchar *unused) 
{
  gchar *name;
  GstPad *sinkpad;
  GstFloat2Int *plugin;

  plugin = GST_FLOAT2INT(element);
  
  g_return_val_if_fail(plugin != NULL, NULL);
  g_return_val_if_fail(GST_IS_FLOAT2INT(plugin), NULL);

  if (templ->direction != GST_PAD_SINK) {
    g_warning ("float2int: request new pad that is not a SINK pad\n");
    return NULL;
  }

  name = g_strdup_printf ("sink%d",plugin->numsinkpads);
  g_print("float2int adding pad %s\n", name);
  sinkpad = gst_pad_new_from_template (templ, name);
  gst_element_add_pad (GST_ELEMENT (plugin), sinkpad);
  gst_pad_set_chain_function(sinkpad,gst_float2int_chain_gint16);
  gst_pad_set_connect_function(sinkpad, gst_float2int_connect);
  sinkpad->element_private = GINT_TO_POINTER(plugin->numsinkpads);
  
  plugin->sinkpads = g_slist_append (plugin->sinkpads, sinkpad);
  plugin->numsinkpads++;
  plugin->channels = plugin->numsinkpads;
  if (plugin->buffers){
    plugin->buffers = g_renew(GstBuffer*,plugin->buffers,plugin->numsinkpads);
    plugin->buffers[plugin->numsinkpads - 1] = NULL;
  }
  else {
    plugin->buffers = g_new0(GstBuffer*,plugin->numsinkpads);
  }
  
  return sinkpad;
}

static void
gst_float2int_chain_gint16 (GstPad *pad,GstBuffer *buf_in)
{
  GstFloat2Int *plugin;
  GstBuffer *buf_out, *buf_temp, **buffers;
  gfloat *data_in;
  gint16 *data_out;
  gint num_frames, i, j;
  guint padpos;
  guint buffer_byte_size;
    
  g_return_if_fail((plugin = gst_float2int_get_plugin(pad,buf_in)));
  
  /* find which position this sinkpad is in */
  buffers = plugin->buffers;
  padpos = GPOINTER_TO_INT(pad->element_private);
  
  if (buffers[padpos]){
    /* g_print("float2int: already have buffer %d\n", padpos); */
    
    /* append the new buffer to the existing one (not ideal, but blame the scheduler :) */
    buffers[padpos] = gst_buffer_append (buffers[padpos], buf_in);
  }
  else {
    /* g_print("float2int: setting buffer %d\n", padpos); */
    buffers[padpos]=buf_in;
    plugin->numbuffers++;
  } 

  /* get the minimum buffer size just in case they vary */
  if (GST_BUFFER_SIZE(buffers[padpos]) < plugin->minbuffersize){
    plugin->minbuffersize = GST_BUFFER_SIZE(buffers[padpos]);
  }
      
  /* check whether we have a full set of buffers yet  */
  if (plugin->numbuffers < plugin->numsinkpads){
    /* g_print("float2int: waiting for %d more buffers\n", plugin->numbuffers); */
    return;
  }
  
  num_frames = plugin->minbuffersize / sizeof(gfloat);
  buffer_byte_size = 2 * num_frames * plugin->channels;
  
  /* g_print("float2int: processing %d frames %d channels\n", num_frames, plugin->channels);  */
  
  buf_out = gst_buffer_new_from_pool(gst_buffer_pool_get_default (buffer_byte_size, 4), 0, 0);
  
  data_out = (gint16*)GST_BUFFER_DATA(buf_out);
  GST_BUFFER_SIZE(buf_out) = buffer_byte_size;
  
  for (i = 0; (i < plugin->channels) && (i < plugin->numsinkpads); i++) {
    
    /* chop the buffer if it is too big */
    if (GST_BUFFER_SIZE(buffers[i]) == plugin->minbuffersize){
      buf_temp = buffers[i];
    }
    else {
      buf_temp =   gst_buffer_create_sub (buffers[i],0,plugin->minbuffersize);
      buffers[i] = gst_buffer_create_sub (buffers[i],plugin->minbuffersize,
                                          GST_BUFFER_SIZE(buffers[i]-plugin->minbuffersize));
    }
    
    data_in=(gfloat *)GST_BUFFER_DATA(buf_temp);
    
    for (j = 0; j < num_frames; j++) {
      
      /* FIXME this is naive, slow and broken - apart from that it works fine */
      data_out[(j*plugin->channels) + i] = (gint16)(data_in[j] * 32767.0);
    }
    
    /* we are done with this buffer */
    gst_buffer_unref(buf_temp);
    if (buf_temp == buffers[i]){
      buffers[i]=NULL;
      plugin->numbuffers--;
    }
    else {
      /* get the new minimum buffer size */
      if (GST_BUFFER_SIZE(buffers[i]) < plugin->minbuffersize){
        plugin->minbuffersize = GST_BUFFER_SIZE(buffers[i]);
      }
    }
  }
  if (!plugin->numbuffers){
    plugin->minbuffersize=G_MAXINT;
  }
  
  gst_pad_push(plugin->srcpad,buf_out);
}

static inline GstFloat2Int* 
gst_float2int_get_plugin(GstPad *pad,GstBuffer *buf)
{
  GstFloat2Int *plugin;
  g_return_val_if_fail(pad != NULL, NULL);
  g_return_val_if_fail(GST_IS_PAD(pad), NULL);
  g_return_val_if_fail(buf != NULL, NULL);

  plugin = GST_FLOAT2INT(GST_OBJECT_PARENT (pad));
  g_return_val_if_fail(plugin != NULL, NULL);
  g_return_val_if_fail(GST_IS_FLOAT2INT(plugin), NULL);
  return plugin;
}

static void
gst_float2int_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
  GstFloat2Int *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FLOAT2INT(object));
  plugin = GST_FLOAT2INT(object);

  switch (prop_id) {
    default:
      break;
  }
}

static void
gst_float2int_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
  GstFloat2Int *plugin;

  /* it's not null if we got it, but it might not be ours */
  g_return_if_fail(GST_IS_FLOAT2INT(object));
  plugin = GST_FLOAT2INT(object);

  switch (prop_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

gboolean 
gst_float2int_factory_init (GstPlugin *plugin) 
{
  GstElementFactory *factory;
  
  factory = gst_element_factory_new("float2int",GST_TYPE_FLOAT2INT,
                                   &float2int_details);
  g_return_val_if_fail(factory != NULL, FALSE);
  
  sinktempl = float2int_sink_factory();
  gst_element_factory_add_pad_template (factory, sinktempl);

  srctempl = float2int_src_factory();
  gst_element_factory_add_pad_template (factory, srctempl);
  
  gst_plugin_add_feature (plugin, GST_PLUGIN_FEATURE (factory));
   
  return TRUE;
}
