/* Pthread-friendly coroutines with pth
 * Copyright (C) 2001 Andy Wingo <wingo@pobox.com>
 *
 * cothread-stack.c: various methods of allocating cothread stacks
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "cothreads-private.h"
#ifdef USE_LINUXTHREADS
# include "linuxthreads.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/resource.h>

typedef enum _cothread_block_state cothread_block_state;
typedef struct _cothread_chunk cothread_chunk;

enum _cothread_block_state
{
  COTHREAD_BLOCK_STATE_UNUSED=0,
  COTHREAD_BLOCK_STATE_IN_USE
};

struct _cothread_chunk {
  cothread_chunk *next;
  cothread_chunk *prev;
  cothread_block_state *block_states;
  char *chunk;
  int size;
  int reserved_bottom;
  gboolean needs_free;
  int nblocks;
};


static cothread_chunk*	cothread_chunk_new		(unsigned long size, gboolean allocate);
static void		cothread_chunk_free		(cothread_chunk *chunk);
static gboolean		cothread_stack_alloc_chunked 	(cothread_chunk *chunk, char **low, char **high, 
                                                         cothread_chunk *(*chunk_new)(cothread_chunk*));
#ifdef USE_LINUXTHREADS
static cothread_chunk*	cothread_chunk_new_linuxthreads	(cothread_chunk* old);
#endif
static cothread_chunk*	cothread_chunk_new_on_stack	(cothread_chunk* old);


gboolean
cothread_stack_alloc_on_heap (char **low, char **high)
{
  if (posix_memalign (low, _cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk,
                      _cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk)) {
    g_error ("could not memalign stack");
    return FALSE;
  }
  
  *high = *low + _cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1;
  return TRUE;
}

#ifdef USE_GTHREADS
gboolean
cothread_stack_alloc_on_gthread_stack (char **low, char **high)
{
  cothread_chunk *chunk = NULL;
  static GStaticPrivate chunk_key = G_STATIC_PRIVATE_INIT;
  
  if (!(chunk = g_static_private_get(&chunk_key))) {
    chunk = cothread_chunk_new (_cothreads_config_global->chunk_size, FALSE);
    g_static_private_set (&chunk_key, chunk, (GDestroyNotify) cothread_chunk_free);
  }
  
  return cothread_stack_alloc_chunked (chunk, low, high, cothread_chunk_new_on_stack);
}
#endif

#ifdef USE_LINUXTHREADS
gboolean
cothread_stack_alloc_linuxthreads (char **low, char **high)
{
  cothread_chunk *chunk = NULL;
  static GStaticPrivate chunk_key = G_STATIC_PRIVATE_INIT;
  
  if (!(chunk = g_static_private_get(&chunk_key))) {
    chunk = cothread_chunk_new (_cothreads_config_global->chunk_size, FALSE);
    g_static_private_set (&chunk_key, chunk, (GDestroyNotify) cothread_chunk_free);
  }
  
  return cothread_stack_alloc_chunked (chunk, low, high, cothread_chunk_new_linuxthreads);
}
#endif

/* size must be a power of two. */
static cothread_chunk*
cothread_chunk_new (unsigned long size, gboolean allocate)
{
  cothread_chunk *ret;
  char *sp = CURRENT_STACK_FRAME;
  
  ret = g_new0 (cothread_chunk, 1);
  ret->nblocks = _cothreads_config_global->blocks_per_chunk;
  ret->block_states = g_new0 (cothread_block_state, ret->nblocks);
  
  if (allocate) {
    if (posix_memalign(&ret->chunk, size, size))
      g_error ("memalign operation failed");
  } else {
    /* if we don't allocate the chunk, we must already be in it. */
    /* set ret->chunk equal to the lower end of the chunk */
    ret->chunk = (char*) ((unsigned long) sp &~ (size - 1));
#if PTH_STACK_GROWTH > 0
    ret->reserved_bottom = sp - ret->chunk + 1;
#else
    ret->reserved_bottom = ret->chunk + size - sp + 1;
#endif
  }
  
  ret->size = size;
  ret->needs_free = allocate;
  
  g_message ("created new chunk, %p, size=0x%x, %s, %d reserved", ret->chunk, ret->size,
             allocate ? "allocated" : "not allocated", ret->reserved_bottom);
  return ret;
}

/**
 * cothread_stack_alloc_chunked:
 * @chunk: the chunk we're working from
 * @low: the lower bound of the stack block passed by reference
 * @high: the upper bound of the stack block passed by reference
 * @chunk_new: how to create a new chunk if this one fills up, or NULL
 *
 * Make a new cothread stack out of a chunk. Chunks are assumed to be aligned on
 * boundaries of _cothreads_config_global->chunk_size.
 *
 * Returns: TRUE if allocation was successful, FALSE otherwise
 */
static gboolean
cothread_stack_alloc_chunked (cothread_chunk *chunk, char **low, char **high, 
                              cothread_chunk *(*chunk_new)(cothread_chunk*))
{
  int block;
  cothread_chunk *walk, *last;
  
  *low = 0;
  *high = 0;
  last = NULL; /* shut up gcc */
  
  for (walk=chunk; walk; last=walk, walk=walk->next) {
    if (walk->block_states[0] == COTHREAD_BLOCK_STATE_UNUSED) {
      walk->block_states[0] = COTHREAD_BLOCK_STATE_IN_USE;
#if PTH_STACK_GROWTH > 0
      *low  = walk->chunk + walk->reserved_bottom;
      *high = walk->chunk + walk->size / walk->nblocks - 1;
#else
      *low  = walk->chunk + walk->size * (walk->nblocks - 1) / walk->nblocks;
      *high = walk->chunk + walk->size - walk->reserved_bottom - 1;
#endif
      return TRUE;
    }
    
    for (block = 1; block < walk->nblocks; block++) {
      if (walk->block_states[block] == COTHREAD_BLOCK_STATE_UNUSED) {
      walk->block_states[block] = COTHREAD_BLOCK_STATE_IN_USE;
#if PTH_STACK_GROWTH > 0
        *low  = walk->chunk + walk->size * block / walk->nblocks;
#else
        *low  = walk->chunk + walk->size * (walk->nblocks - block - 1) / walk->nblocks;
#endif
        *high = *low + walk->size / walk->nblocks - 1;
        
        return TRUE;
      }
    }
  }
  
  if (!chunk_new)
    return FALSE;
  else
    return cothread_stack_alloc_chunked (chunk_new (last), low, high, NULL);
}

static void
cothread_chunk_free (cothread_chunk *chunk) 
{
  /* FIXME: implement me please */
}

#ifdef USE_LINUXTHREADS
static cothread_chunk*
cothread_chunk_new_linuxthreads (cothread_chunk* old)
{
  cothread_chunk *new, *walk;
  void *descr;
  
  descr = __linuxthreads_self();
  
  /* if we're in the first thread */
  for (walk=old; walk; walk=walk->prev)
    if (descr >= walk->chunk && descr < walk->chunk + walk->size)
      break;
  if (!walk)
    return cothread_chunk_new_on_stack (old);
    
  new = cothread_chunk_new (_cothreads_config_global->chunk_size, TRUE);
  
#if PTH_STACK_GROWTH > 0
  /* we don't really know descr's size in this case, but we can be
   * conservative. it's normally 1K in the down-growing case. we have to
   * allocate at least one page's worth, though.
   */
  new->reserved_bottom = getpagesize();
#ifdef COTHREADS_DEBUG
  g_message ("copying reserved_bottom %d from %p to %p", new->reserved_bottom, descr, new->chunk);
#endif
  memcpy(new->chunk, descr, new->reserved_bottom);
#else
/*  new->reserved_bottom = ((unsigned long) descr | (new->size - 1)) - (unsigned long) descr + 1; */
  new->reserved_bottom = getpagesize();
#ifdef COTHREADS_DEBUG
  g_message ("copying reserved_bottom %d from %p to %p", new->reserved_bottom,
             old->chunk + old->size - new->reserved_bottom, 
             new->chunk + new->size - new->reserved_bottom);
#endif
  memcpy(new->chunk + new->size - new->reserved_bottom,
         old->chunk + old->size - new->reserved_bottom, new->reserved_bottom);
#endif
  g_message ("new chunk: reserved linuxthreads = %d", new->reserved_bottom);
  
  old->next = new;
  new->prev = old;
  return new;
}
#endif

static cothread_chunk*
cothread_chunk_new_on_stack (cothread_chunk* old)
{
  cothread_chunk *new;
  struct rlimit limit;
  
  getrlimit (RLIMIT_STACK, &limit);
#ifdef COTHREADS_DEBUG
  g_print ("stack limit: %ld\nstack max: %ld\n", limit.rlim_cur, limit.rlim_max);
#endif
  limit.rlim_cur += old->size;
  if (setrlimit (RLIMIT_STACK, &limit)) {
    perror ("Could not increase the stack size, aborting...");
    return NULL;
  }
  
  new = cothread_chunk_new (old->size, FALSE);
  new->reserved_bottom = 0;
  
#if PTH_STACK_GROWTH > 0
  new->chunk = old->chunk + old->size;
#else
  new->chunk = old->chunk - old->size;
#endif
  g_message ("new chunk allocated on the stack");
  
  old->next = new;
  new->prev = old;
  return new;
}
