/* Pthread-friendly coroutines with pth
 * Copyright (C) 2002 Andy Wingo <wingo@pobox.com>
 *
 * cothreads.c: public API implementation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "pth_p.h" /* this pulls in everything */
#include <sys/mman.h>

#ifdef USE_LINUXTHREADS
static cothreads_config cothreads_config_default = COTHREADS_CONFIG_LINUXTHREADS_INITIALIZER;
#else
#ifdef USE_GTHREADS
static cothreads_config cothreads_config_default = COTHREADS_CONFIG_GTHREAD_INITIALIZER;
#else
static cothreads_config cothreads_config_default = COTHREADS_CONFIG_HEAP_INITIALIZER;
#endif
#endif

cothreads_config *_cothreads_config_global = NULL;

static gboolean (*stack_alloc_func) (char**, char**);
static gboolean (*stack_alloc_func_0) (char**, char**);

static void*	cothread_private_set	(char *sp, void *priv, size_t size);
static void*	cothread_private_get	(char *sp, void *priv, size_t size);
static void	cothread_stub		(void);


/**
 * cothreads_initialized:
 *
 * Query the state of the cothreads system.
 *
 * Returns: TRUE if cothreads_init() has already been called, FALSE otherwise
 */
gboolean
cothreads_initialized (void) 
{
  return (_cothreads_config_global != NULL);
}

/**
 * cothreads_init:
 * @config: attributes for creation of cothread stacks
 *
 * Initialize the cothreads system. If @config is NULL, use the default parameters
 * detected at compile-time.
 */
void
cothreads_init (cothreads_config *config)
{
  static cothreads_config _config;
  
  if (cothreads_initialized()) {
    g_warning ("cothread system has already been initialized");
    return;
  }

  /* we don't hold on to *config, we copy it (if it's supplied) */
  if (!config)
    _config = cothreads_config_default;
  else 
    _config = *config;
  
  _cothreads_config_global = &_config;
  
  switch (_cothreads_config_global->method) {
  case COTHREADS_ALLOC_METHOD_MALLOC:
    g_message ("Initializing cothreads routine, heap stack allocation method");
    stack_alloc_func_0 = NULL;
    stack_alloc_func = cothread_stack_alloc_on_heap;
    break;
#ifdef USE_GTHREADS
  case COTHREADS_ALLOC_METHOD_GTHREAD_STACK:
    g_message ("Initializing cothreads routine, gthreads stack allocation method");
    stack_alloc_func_0 = cothread_stack_alloc_on_gthread_stack;
    stack_alloc_func = cothread_stack_alloc_on_gthread_stack;
    break;
#endif
#ifdef USE_LINUXTHREADS
  case COTHREADS_ALLOC_METHOD_LINUXTHREADS:
    g_message ("Initializing cothreads routine, linuxthreads stack allocation method");
    stack_alloc_func_0 = cothread_stack_alloc_linuxthreads;
    stack_alloc_func = cothread_stack_alloc_linuxthreads;
    break;
#endif
  default:
    g_error ("unexpected value for config method %d", _cothreads_config_global->method);
  }
  g_message ("current stack pointer: %p", CURRENT_STACK_FRAME);
}

/**
 * cothread_create:
 * @func: function to start with this cothread
 * @argc: argument count
 * @argv: argument vector
 * @main: cothread to switch back to if @func exits, or NULL
 *
 * Create a new cothread running a given function. You must explictly switch
 * into this cothread to give it control. If @func is NULL, a cothread is
 * created on the current stack with the current stack pointer.
 *
 * Returns: A pointer to the new cothread
 */
cothread*
cothread_create (cothread_func func, int argc, void **argv, cothread *main)
{
  char *low, *high, *dest;
  cothread_private priv;
  cothread *ret = NULL;
  
  low = high = NULL;
  memset (&priv, 0, sizeof (priv));
  
  if (!func) {
    /* we are being asked to save the current thread into a new cothread. this
     * only happens for the first cothread in a thread. */
    low = CURRENT_STACK_FRAME;
#if PTH_STACK_GROWTH > 0
    dest = (char*) ((gulong)low | (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
      + 1 - getpagesize() * 2;
#else
    dest = (char*) ((gulong)low &~ (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
      + getpagesize();
#endif
#ifdef COTHREADS_DEBUG
    g_message ("about to mmap %p, size %d", dest, getpagesize());
#endif
    if (mmap (dest, getpagesize(), PROT_READ|PROT_WRITE,
              MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0) == MAP_FAILED) {
      g_critical ("mmap failed, captain");
      return NULL;
    }
    ret = (cothread*) cothread_private_set (low, &priv, sizeof(priv));
    if (stack_alloc_func_0 && !stack_alloc_func_0 (&low, &high))
      g_error ("couldn't create cothread 0");
    else
      g_message ("created cothread 0 (%p) with low=%p, high=%p", ret, low, high);
    
    pth_mctx_save (ret);
    return ret;
  }
  
  if (!stack_alloc_func (&low, &high))
    g_error ("could not allocate a new cothread stack");
  
#ifdef COTHREADS_DEBUG
  g_message ("about to mmap %p, size %d", low, high - low + 1);
#endif
  if (mmap (low, high - low + 1, PROT_EXEC|PROT_READ|PROT_WRITE,
            MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0) == MAP_FAILED) {
    g_critical ("mmap failed, captain");
    return NULL;
  }
  
  priv.argc = argc;
  priv.argv = argv;
  priv.func = func;
  priv.main = main;
  ret = (cothread*)cothread_private_set (low, &priv, sizeof(priv));
  
  g_message ("created a cothread %p with low=%p, high=%p", ret, low, high);
  
  pth_mctx_set (ret, cothread_stub, low, high);
  
  return ret;
}

/**
 * cothread_self:
 *
 * Gets the current cothread.
 *
 * Returns: A pointer to the current cothread
 */
cothread*
cothread_self (void)
{
  char *sp = CURRENT_STACK_FRAME;

  return (cothread*)cothread_private_get (sp, NULL, sizeof (cothread_private));
}

/**
 * cothread_reset:
 * @self: the cothread, created by cothread_create
 * @func: function to start with this cothread
 * @argc: argument count
 * @argv: argument vector
 * @main: cothread to switch back to if @func exits, or NULL
 *
 * Reset a cothread to start running @func, as if it were newly created.
 */
void
cothread_reset (cothread *self, cothread_func func, int argc, void **argv, cothread *main)
{
  char *low, *high;
  cothread_private priv;
  
  g_return_if_fail (self != NULL);

  low = (char*) ((gulong)self &~ (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1));
  high = (char*) ((gulong)self | (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1));

  pth_mctx_set (self, cothread_stub, low, high);

  priv.self = *self;
  priv.argc = argc;
  priv.argv = argv;
  priv.func = func;
  priv.main = main;
  cothread_private_set (low, &priv, sizeof(priv));
}
  
/**
 * cothread_destroy:
 * @thread: cothread to destroy
 *
 * Deallocate any memory used by the cothread data structures.
 */
void
cothread_destroy (cothread *thread)
{
  gchar *low, *dest;
  /* FIXME: have method-specific destroy functions. */
  
  /* FIXME: this is hackish */

  /* unmap the private data */
  low = CURRENT_STACK_FRAME;
#if PTH_STACK_GROWTH > 0
  dest = (char*) ((gulong)low | (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    + 1 - getpagesize() * 2;
#else
  dest = (char*) ((gulong)low &~ (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    + getpagesize();
#endif
/*  if (munmap (dest, getpagesize()) != 0)
    perror ("munmap failed"); */
}

/* the whole 'page size' thing is to allow for the last page of a stack or chunk
 * to be mmap'd as a boundary page */

static void*
cothread_private_set (char *sp, void *priv, size_t size)
{
  char *dest;
  
#if PTH_STACK_GROWTH > 0
  dest = (char*) ((gulong)sp | (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    - size + 1 - getpagesize();
#else
  dest = (char*) ((gulong)sp &~ (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    + getpagesize();
#endif
  
#ifdef COTHREADS_DEBUG
  g_message ("setting %p to equal %p, size %d", dest, priv, size);
#endif
  memcpy (dest, priv, size);
  return dest;
}

static void*
cothread_private_get (char *sp, void *priv, size_t size)
{
  char *src;
  
#if PTH_STACK_GROWTH > 0
  src = (char*) ((gulong)sp | (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    - size + 1 - getpagesize();
#else
  src = (char*) ((gulong)sp &~ (_cothreads_config_global->chunk_size / _cothreads_config_global->blocks_per_chunk - 1))
    + getpagesize();
#endif
  
  if (priv)
    memcpy (priv, src, size);

  return src;
}

static void
cothread_stub (void)
{
  cothread_private *priv;
  char *sp = CURRENT_STACK_FRAME;
  
  priv = (cothread_private*) cothread_self();

  /* this idiom is stupid but it's what the old cothreads does. to quote:
   * "we do this to avoid ever returning, we just switch to 0th thread"
   */
  while (TRUE) {
    priv->func (priv->argc, priv->argv);
    
    if (priv->main) {
#ifdef COTHREADS_DEBUG
      g_message ("returning from cothread %p to %p", &priv->self, priv->main);
#endif
      cothread_switch (&priv->self, priv->main);
    }
  }
  
  g_critical ("returning from cothread_stub!");
}

